@@ -16,11 +16,12 @@ use 5.006;
 use strict;
 use warnings;
 
-use Carp qw< carp croak >;
+use Carp         qw< carp croak >;
+use Scalar::Util qw< blessed >;
 
-use Math::BigFloat 1.999718;
+use Math::BigFloat ();
 
-our $VERSION = '0.2616';
+our $VERSION = '0.2618';
 
 our @ISA = qw(Math::BigFloat);
 
@@ -366,6 +367,11 @@ sub new {
 
     # At this point, neither $n nor $d is a NaN or a zero.
 
+    # Copy them now before manipulating them.
+
+    $n = $n -> copy();
+    $d = $d -> copy();
+
     if ($d < 0) {               # make sure denominator is positive
         $n -> bneg();
         $d -> bneg();
@@ -788,12 +794,21 @@ sub badd {
         ($class, $x, $y, @r) = objectify(2, @_);
     }
 
-    # +inf + +inf => +inf, -inf + -inf => -inf
-    return $x->binf(substr($x->{sign}, 0, 1))
-      if $x->{sign} eq $y->{sign} && $x->{sign} =~ /^[+-]inf$/;
-
-    # +inf + -inf or -inf + +inf => NaN
-    return $x->bnan() if ($x->{sign} !~ /^[+-]$/ || $y->{sign} !~ /^[+-]$/);
+    unless ($x -> is_finite() && $y -> is_finite()) {
+        if ($x -> is_nan() || $y -> is_nan()) {
+            return $x -> bnan(@r);
+        } elsif ($x -> is_inf("+")) {
+            return $x -> bnan(@r) if $y -> is_inf("-");
+            return $x -> binf("+", @r);
+        } elsif ($x -> is_inf("-")) {
+            return $x -> bnan(@r) if $y -> is_inf("+");
+            return $x -> binf("-", @r);
+        } elsif ($y -> is_inf("+")) {
+            return $x -> binf("+", @r);
+        } elsif ($y -> is_inf("-")) {
+            return $x -> binf("-", @r);
+        }
+    }
 
     #  1   1    gcd(3, 4) = 1    1*3 + 1*4    7
     #  - + -                  = --------- = --
@@ -1110,6 +1125,20 @@ sub binc {
     $x->bnorm()->round(@r);
 }
 
+sub binv {
+    my $x = shift;
+    my @r = @_;
+
+    return $x if $x->modify('binv');
+
+    return $x              if $x -> is_nan();
+    return $x -> bzero()   if $x -> is_inf();
+    return $x -> binf("+") if $x -> is_zero();
+
+    ($x -> {_n}, $x -> {_d}) = ($x -> {_d}, $x -> {_n});
+    $x -> round(@r);
+}
+
 ##############################################################################
 # is_foo methods (the rest is inherited)
 
@@ -1201,6 +1230,35 @@ sub parts {
     ($n, $d);
 }
 
+sub dparts {
+    my $x = shift;
+    my $class = ref $x;
+
+    croak("dparts() is an instance method") unless $class;
+
+    if ($x -> is_nan()) {
+        return $class -> bnan(), $class -> bnan() if wantarray;
+        return $class -> bnan();
+    }
+
+    if ($x -> is_inf()) {
+        return $class -> binf($x -> sign()), $class -> bzero() if wantarray;
+        return $class -> binf($x -> sign());
+    }
+
+    # 355/113 => 3 + 16/113
+
+    my ($q, $r)  = $LIB -> _div($LIB -> _copy($x -> {_n}), $x -> {_d});
+
+    my $int = Math::BigRat -> new($x -> {sign} . $LIB -> _str($q));
+    return $int unless wantarray;
+
+    my $frc = Math::BigRat -> new($x -> {sign} . $LIB -> _str($r),
+                                  $LIB -> _str($x -> {_d}));
+
+    return $int, $frc;
+}
+
 sub length {
     my ($class, $x) = ref($_[0]) ? (undef, $_[0]) : objectify(1, @_);
 
@@ -1244,7 +1302,7 @@ sub bfloor {
 }
 
 sub bint {
-    my ($class, $x, @r) = ref($_[0]) ? (ref($_[0]), $_[0]) : objectify(1, @_);
+    my ($class, $x) = ref($_[0]) ? (ref($_[0]), $_[0]) : objectify(1, @_);
 
     return $x if ($x->{sign} !~ /^[+-]$/ ||     # +/-inf or NaN
                   $LIB -> _is_one($x->{_d}));   # already an integer
@@ -1348,11 +1406,11 @@ sub blog {
     # $x->blog(undef) signals that the base is Euler's number.
 
     if (!ref($_[0]) && $_[0] =~ /^[A-Za-z]|::/) {
-        # E.g., Math::BigFloat->blog(256, 2)
+        # E.g., Math::BigRat->blog(256, 2)
         ($class, $x, $base, @r) =
           defined $_[2] ? objectify(2, @_) : objectify(1, @_);
     } else {
-        # E.g., Math::BigFloat::blog(256, 2) or $x->blog(2)
+        # E.g., Math::BigRat::blog(256, 2) or $x->blog(2)
         ($class, $x, $base, @r) =
           defined $_[1] ? objectify(2, @_) : objectify(1, @_);
     }
@@ -1393,6 +1451,24 @@ sub blog {
         return $x -> binf($sign);
     }
 
+    # Now take care of the cases where $x and/or $base is 1/N.
+    #
+    #   log(1/N) / log(B)   = -log(N)/log(B)
+    #   log(1/N) / log(1/B) =  log(N)/log(B)
+    #   log(N)   / log(1/B) = -log(N)/log(B)
+
+    my $neg = 0;
+    if ($x -> numerator() -> is_one()) {
+        $x -> binv();
+        $neg = !$neg;
+    }
+    if (defined(blessed($base)) && $base -> isa($class)) {
+        if ($base -> numerator() -> is_one()) {
+            $base = $base -> copy() -> binv();
+            $neg = !$neg;
+        }
+    }
+
     # At this point we are done handling all exception cases and trivial cases.
 
     $base = Math::BigFloat -> new($base) if defined $base;
@@ -1406,7 +1482,7 @@ sub blog {
     $x -> {_n}   = $xtmp -> {_n};
     $x -> {_d}   = $xtmp -> {_d};
 
-    return $x;
+    return $neg ? $x -> bneg() : $x;
 }
 
 sub bexp {
@@ -1575,7 +1651,7 @@ sub bmodpow {
     my $yint = Math::BigInt -> new($y -> copy() -> bint());
     my $mint = Math::BigInt -> new($m -> copy() -> bint());
 
-    $xint -> bmodpow($y, $m, @r);
+    $xint -> bmodpow($yint, $mint, @r);
     my $xtmp = Math::BigRat -> new($xint -> bsstr());
 
     $x -> {sign} = $xtmp -> {sign};
@@ -1597,7 +1673,7 @@ sub bmodinv {
     my $xint = Math::BigInt -> new($x -> copy() -> bint());
     my $yint = Math::BigInt -> new($y -> copy() -> bint());
 
-    $xint -> bmodinv($y, @r);
+    $xint -> bmodinv($yint, @r);
     my $xtmp = Math::BigRat -> new($xint -> bsstr());
 
     $x -> {sign} = $xtmp -> {sign};
@@ -1656,7 +1732,7 @@ sub bsqrt {
 }
 
 sub blsft {
-    my ($class, $x, $y, $b, @r) = objectify(2, @_);
+    my ($class, $x, $y, $b) = objectify(2, @_);
 
     $b = 2 if !defined $b;
     $b = $class -> new($b) unless ref($b) && $b -> isa($class);
@@ -1670,7 +1746,7 @@ sub blsft {
 }
 
 sub brsft {
-    my ($class, $x, $y, $b, @r) = objectify(2, @_);
+    my ($class, $x, $y, $b) = objectify(2, @_);
 
     $b = 2 if !defined $b;
     $b = $class -> new($b) unless ref($b) && $b -> isa($class);
@@ -1808,9 +1884,9 @@ sub bcmp {
 
     if ($x->{sign} !~ /^[+-]$/ || $y->{sign} !~ /^[+-]$/) {
         # $x is NaN and/or $y is NaN
-        return undef if $x->{sign} eq $nan || $y->{sign} eq $nan;
+        return       if $x->{sign} eq $nan || $y->{sign} eq $nan;
         # $x and $y are both either +inf or -inf
-        return 0     if $x->{sign} eq $y->{sign} && $x->{sign} =~ /^[+-]inf$/;
+        return  0    if $x->{sign} eq $y->{sign} && $x->{sign} =~ /^[+-]inf$/;
         # $x = +inf and $y < +inf
         return +1    if $x->{sign} eq '+inf';
         # $x = -inf and $y > -inf
@@ -1855,9 +1931,9 @@ sub bacmp {
 
     if (($x->{sign} !~ /^[+-]$/) || ($y->{sign} !~ /^[+-]$/)) {
         # handle +-inf and NaN
-        return undef if (($x->{sign} eq $nan) || ($y->{sign} eq $nan));
-        return 0 if $x->{sign} =~ /^[+-]inf$/ && $y->{sign} =~ /^[+-]inf$/;
-        return 1 if $x->{sign} =~ /^[+-]inf$/ && $y->{sign} !~ /^[+-]inf$/;
+        return    if (($x->{sign} eq $nan) || ($y->{sign} eq $nan));
+        return  0 if $x->{sign} =~ /^[+-]inf$/ && $y->{sign} =~ /^[+-]inf$/;
+        return  1 if $x->{sign} =~ /^[+-]inf$/ && $y->{sign} !~ /^[+-]inf$/;
         return -1;
     }
 
@@ -1869,7 +1945,7 @@ sub bacmp {
 sub beq {
     my $self    = shift;
     my $selfref = ref $self;
-    my $class   = $selfref || $self;
+    #my $class   = $selfref || $self;
 
     croak 'beq() is an instance method, not a class method' unless $selfref;
     croak 'Wrong number of arguments for beq()' unless @_ == 1;
@@ -1881,7 +1957,7 @@ sub beq {
 sub bne {
     my $self    = shift;
     my $selfref = ref $self;
-    my $class   = $selfref || $self;
+    #my $class   = $selfref || $self;
 
     croak 'bne() is an instance method, not a class method' unless $selfref;
     croak 'Wrong number of arguments for bne()' unless @_ == 1;
@@ -1893,7 +1969,7 @@ sub bne {
 sub blt {
     my $self    = shift;
     my $selfref = ref $self;
-    my $class   = $selfref || $self;
+    #my $class   = $selfref || $self;
 
     croak 'blt() is an instance method, not a class method' unless $selfref;
     croak 'Wrong number of arguments for blt()' unless @_ == 1;
@@ -1905,7 +1981,7 @@ sub blt {
 sub ble {
     my $self    = shift;
     my $selfref = ref $self;
-    my $class   = $selfref || $self;
+    #my $class   = $selfref || $self;
 
     croak 'ble() is an instance method, not a class method' unless $selfref;
     croak 'Wrong number of arguments for ble()' unless @_ == 1;
@@ -1917,7 +1993,7 @@ sub ble {
 sub bgt {
     my $self    = shift;
     my $selfref = ref $self;
-    my $class   = $selfref || $self;
+    #my $class   = $selfref || $self;
 
     croak 'bgt() is an instance method, not a class method' unless $selfref;
     croak 'Wrong number of arguments for bgt()' unless @_ == 1;
@@ -1929,7 +2005,7 @@ sub bgt {
 sub bge {
     my $self    = shift;
     my $selfref = ref $self;
-    my $class   = $selfref || $self;
+    #my $class   = $selfref || $self;
 
     croak 'bge() is an instance method, not a class method'
         unless $selfref;
@@ -2024,18 +2100,27 @@ sub as_oct {
 sub from_hex {
     my $class = shift;
 
-    $class->new(@_);
+    # The relationship should probably go the otherway, i.e, that new() calls
+    # from_hex(). Fixme!
+    my ($x, @r) = @_;
+    $x =~ s|^\s*(?:0?[Xx]_*)?|0x|;
+    $class->new($x, @r);
 }
 
 sub from_bin {
     my $class = shift;
 
-    $class->new(@_);
+    # The relationship should probably go the otherway, i.e, that new() calls
+    # from_bin(). Fixme!
+    my ($x, @r) = @_;
+    $x =~ s|^\s*(?:0?[Bb]_*)?|0b|;
+    $class->new($x, @r);
 }
 
 sub from_oct {
     my $class = shift;
 
+    # Why is this different from from_hex() and from_bin()? Fixme!
     my @parts;
     for my $c (@_) {
         push @parts, Math::BigInt->from_oct($c);
@@ -2048,53 +2133,82 @@ sub from_oct {
 
 sub import {
     my $class = shift;
-    my @a;
-    my $lib = '';
-    my $try = 'try';
+    my @a;                      # unrecognized arguments
+    my $lib_param = '';
+    my $lib_value = '';
 
-    for (my $i = 0; $i <= $#_ ; $i++) {
-        croak "Error in import(): argument with index $i is undefined"
-          unless defined($_[$i]);
+    while (@_) {
+        my $param = shift;
 
-        if ($_[$i] eq ':constant') {
-            # this rest causes overlord er load to step in
+        # Enable overloading of constants.
+
+        if ($param eq ':constant') {
             overload::constant float => sub { $class->new(shift); };
+            next;
         }
 
-        #elsif ($_[$i] eq 'upgrade') {
-        #    # this causes upgrading
-        #    $upgrade = $_[$i+1];        # or undef to disable
-        #    $i++;
-        #}
+        # Upgrading.
 
-        elsif ($_[$i] eq 'downgrade') {
-            # this causes downgrading
-            $downgrade = $_[$i+1];      # or undef to disable
-            $i++;
+        if ($param eq 'upgrade') {
+            $class -> upgrade(shift);
+            next;
         }
 
-        elsif ($_[$i] =~ /^(lib|try|only)\z/) {
-            $lib = $_[$i+1] || '';
-            $try = $1;                  # "lib", "try" or "only"
-            $i++;
+        # Downgrading.
+
+        if ($param eq 'downgrade') {
+            $class -> downgrade(shift);
+            next;
         }
 
-        elsif ($_[$i] eq 'with') {
-            # this argument is no longer used
-            # $LIB = $_[$i+1] || 'Calc';
+        # Accuracy.
+
+        if ($param eq 'accuracy') {
+            $class -> accuracy(shift);
+            next;
+        }
+
+        # Precision.
+
+        if ($param eq 'precision') {
+            $class -> precision(shift);
+            next;
+        }
+
+        # Rounding mode.
+
+        if ($param eq 'round_mode') {
+            $class -> round_mode(shift);
+            next;
+        }
+
+        # Backend library.
+
+        if ($param =~ /^(lib|try|only)\z/) {
+            # alternative library
+            $lib_param = $param;        # "lib", "try", or "only"
+            $lib_value = shift;
+            next;
+        }
+
+        if ($param eq 'with') {
+            # alternative class for our private parts()
+            # XXX: no longer supported
+            # $LIB = shift() || 'Calc';
             # carp "'with' is no longer supported, use 'lib', 'try', or 'only'";
-            $i++;
+            shift;
+            next;
         }
 
-        else {
-            push @a, $_[$i];
-        }
+        # Unrecognized parameter.
+
+        push @a, $param;
     }
 
     require Math::BigInt;
 
     my @import = ('objectify');
-    push @import, $try, $lib if $lib ne '';
+    push @import, $lib_param, $lib_value if $lib_param ne '';
     Math::BigInt -> import(@import);
 
     # find out which one was actually loaded
@@ -2206,6 +2320,10 @@ Returns a copy of the denominator (the part under the line) as positive BigInt.
 Return a list consisting of (signed) numerator and (unsigned) denominator as
 BigInts.
 
+=item dparts()
+
+Returns the integer part and the fraction part.
+
 =item numify()
 
     my $y = $x->numify();
@@ -2520,6 +2638,12 @@ does floored division (F-division), returning an integer $q and a remainder $r
 so that $x = $q * $y + $r. The remainer (modulo) is equal to what is returned
 by C<< $x->bmod($y) >>.
 
+=item binv()
+
+    $x->binv();
+
+Inverse of $x.
+
 =item bdec()
 
     $x->bdec();
